import json
import logging
from collections import OrderedDict, defaultdict
from enum import Enum
from errno import ENOENT
from os import strerror
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pyarrow as pa
import ray
import s3fs
from pyarrow import parquet as pq
from pyarrow.fs import FileSystem, FileType, S3FileSystem
from ray.data.block import Block, BlockMetadata
from ray.data.datasource import (
    BlockWritePathProvider,
    CSVDatasource,
    DefaultBlockWritePathProvider,
    DefaultFileMetadataProvider,
    ParquetBaseDatasource,
    ParquetMetadataProvider,
    PathPartitionParser,
)
from ray.data.datasource.datasource import ArrowRow, Datasource, ReadTask, WriteResult
from ray.data.datasource.file_based_datasource import _resolve_paths_and_filesystem
from ray.data.datasource.file_meta_provider import FastFileMetadataProvider
from ray.types import ObjectRef

from deltacat import ContentEncoding, ContentType, logs
from deltacat.aws.redshift.model.manifest import (
    Manifest,
    ManifestEntry,
    ManifestEntryList,
    ManifestMeta,
)
from deltacat.aws.s3u import (
    S3Url,
    filter_objects_by_prefix,
    objects_to_paths,
    parse_s3_url,
)
from deltacat.types.media import DELIMITED_TEXT_CONTENT_TYPES
from deltacat.utils.common import ReadKwargsProvider

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class CapturingBlockWritePathProvider(BlockWritePathProvider):
    """Delegating block write path provider that saves an ordered dictionary of
    input keyword arguments for every block write path returned."""

    def __init__(self, block_write_path_provider: BlockWritePathProvider):
        self.block_write_path_provider = block_write_path_provider
        self.write_path_kwargs: Dict[str, Dict[str, Any]] = OrderedDict()

    def _get_write_path_for_block(self, base_path: str, *args, **kwargs) -> str:
        write_path = self.block_write_path_provider(
            base_path,
            *args,
            **kwargs,
        )
        kwargs["base_path"] = base_path
        self.write_path_kwargs[write_path] = kwargs
        return write_path


class CachedFileMetadataProvider(
    FastFileMetadataProvider,
    ParquetMetadataProvider,
):
    def __init__(self, meta_cache: Dict[str, BlockMetadata]):
        self._meta_cache = meta_cache

    def get_meta_cache(self) -> Dict[str, BlockMetadata]:
        return self._meta_cache

    def _get_block_metadata(
        self,
        paths: List[str],
        schema: Optional[Union[type, pa.Schema]],
        **kwargs,
    ) -> BlockMetadata:
        agg_block_metadata = BlockMetadata(
            num_rows=0,
            size_bytes=0,
            schema=schema,
            input_files=[],
            exec_stats=None,
        )
        for path in paths:
            block_metadata = self._meta_cache.get(path)
            if block_metadata is None:
                raise ValueError(f"Block metadata not found for path: {path}")
            if block_metadata.num_rows is None:
                agg_block_metadata.num_rows = None
            elif agg_block_metadata.num_rows is not None:
                agg_block_metadata.num_rows += block_metadata.num_rows
            if block_metadata.size_bytes is None:
                agg_block_metadata.size_bytes = None
            elif agg_block_metadata.size_bytes is not None:
                agg_block_metadata.size_bytes += block_metadata.size_bytes
            agg_block_metadata.input_files.append(path)
        return agg_block_metadata


class HivePartitionParser(PathPartitionParser):
    def __init__(
        self,
        base_dir: Optional[str] = None,
        filter_fn: Optional[Callable[[Dict[str, str]], bool]] = None,
    ):
        super(HivePartitionParser, self).__init__(
            base_dir=base_dir,
            filter_fn=filter_fn,
        )


class RedshiftUnloadTextArgs:
    def __init__(
        self,
        csv: bool = False,
        header: bool = False,
        delimiter: Optional[str] = None,
        bzip2: bool = False,
        gzip: bool = False,
        zstd: bool = False,
        add_quotes: Optional[bool] = None,
        null_as: str = "",
        escape: bool = False,
        fixed_width: bool = False,
    ):
        self.header = header
        self.delimiter = delimiter if delimiter else "," if csv else "|"
        self.bzip2 = bzip2
        self.gzip = gzip
        self.zstd = zstd
        self.add_quotes = add_quotes if add_quotes else True if csv else False
        self.null_as = null_as
        self.escape = escape
        self.fixed_width = fixed_width

    def _get_arrow_compression_codec_name(self) -> str:
        arrow_compression_codec_name = None
        codecs_enabled = {
            "bz2": self.bzip2,
            "gzip": self.gzip,
            "zstd": self.zstd,
        }
        for encoding, flag in codecs_enabled.items():
            if arrow_compression_codec_name and flag:
                raise ValueError(
                    f"Multiple Redshift UNLOAD compression types specified "
                    f"({codecs_enabled}). Please ensure that only one "
                    f"compression type is set and try again."
                )
            if flag:
                arrow_compression_codec_name = encoding
        return arrow_compression_codec_name

    def to_arrow_reader_kwargs(
        self, include_columns: Optional[List[str]], schema: Optional[pa.Schema]
    ) -> Dict[str, Any]:
        from pyarrow import csv

        if self.fixed_width:
            raise NotImplementedError(
                "Redshift text files unloaded with FIXEDWIDTH are not "
                "currently supported."
            )
        open_stream_args = {}
        arrow_compression_codec_name = self._get_arrow_compression_codec_name()
        if arrow_compression_codec_name:
            open_stream_args["compression"] = arrow_compression_codec_name
        column_names = None
        if schema:
            column_names = schema.names
        autogen_column_names = False if self.header or column_names else True
        read_options = csv.ReadOptions(
            use_threads=False,
            column_names=column_names,
            autogenerate_column_names=autogen_column_names,
        )
        parse_options = csv.ParseOptions(
            delimiter=self.delimiter,
            quote_char='"' if self.add_quotes else False,
            escape_char="\\" if self.escape else False,
            double_quote=False if self.escape else True,
        )
        convert_options = csv.ConvertOptions(
            column_types=schema,
            null_values=[self.null_as] if self.null_as is not None else [],
            true_values=["t"],
            false_values=["f"],
            strings_can_be_null=True if self.null_as is not None else False,
            quoted_strings_can_be_null=True if self.null_as else False,
            include_columns=include_columns,
        )
        return {
            "open_stream_args": open_stream_args,
            "read_options": read_options,
            "parse_options": parse_options,
            "convert_options": convert_options,
        }


class S3PathType(str, Enum):
    MANIFEST = "manifest"
    PREFIX = "prefix"
    FILES_AND_FOLDERS = "files_and_folders"


class RedshiftWriteResult:
    def __init__(self):
        self.metadata = None
        self.path = None
        self.dataset_uuid = None
        self.block_write_path_provider = None
        self.content_type = None
        self.content_encoding = None
        self.filesystem = None


def _normalize_s3_paths_for_filesystem(
    paths: Union[str, List[str]],
    filesystem: Union[S3FileSystem, s3fs.S3FileSystem],
) -> Tuple[List[str], List[S3Url]]:
    if isinstance(paths, str):
        paths = [paths]
    urls = [parse_s3_url(url) for url in paths]
    if isinstance(filesystem, FileSystem):
        # pyarrow.fs.FileSystem paths should not start with "s3://"
        # pyarrow.fs.FileSystem paths should not end with "/"
        paths = [f"{u.bucket}/{u.key}".rstrip("/") for u in urls]
    else:
        # s3fs.S3FileSystem can start with "s3://" (presumably others can too)
        paths = [u.url.rstrip("/") for u in urls]
    return paths, urls


def _read_manifest_entry_paths(
    entries: ManifestEntryList,
    manifest_content_type: Optional[str],
    content_type_provider: Callable[[str], ContentType],
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    # support manifests with heterogenous content types
    content_type_to_paths = defaultdict(list)
    meta_cache: Dict[str, BlockMetadata] = {}
    for e in entries:
        url = e.url if e.url else e.uri
        # get manifest entry content type or fall back to manifest content type
        content_type = e.meta.content_type or manifest_content_type
        if content_type:
            content_type_to_paths[ContentType(content_type)] = url
        else:
            # fall back to content type inference by file extension
            content_type_to_paths[content_type_provider(url)].append(url)
        meta_cache[url] = BlockMetadata(
            num_rows=e.meta.record_count,
            size_bytes=e.meta.content_length,
            schema=None,
            input_files=[],
            exec_stats=None,
        )
    return content_type_to_paths, CachedFileMetadataProvider(meta_cache)


def _expand_manifest_paths(
    paths: List[str],
    filesystem: Optional[Union[S3FileSystem, s3fs.S3FileSystem]],
    content_type_provider: Callable[[str], ContentType],
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    assert len(paths) == 1, f"Expected 1 manifest path, found {len(paths)}."
    path = paths[0]
    with filesystem.open_input_file(path) as f:
        manifest = Manifest(json.loads(f.read()))
    content_type_to_paths = {}
    meta_provider = CachedFileMetadataProvider({})
    if not manifest.entries:
        logger.warning(f"No entries to read in Redshift Manifest: {path}")
    else:
        content_type_to_paths, meta_provider = _read_manifest_entry_paths(
            manifest.entries,
            manifest.meta.content_type if manifest.meta else None,
            content_type_provider,
        )
    # TODO(pdames): infer the schema from a verbose manifest if available?
    # if not schema and ContentType.PARQUET not in content_type_to_paths:
    #     schema = _infer_schema_from_manifest(manifest)
    return content_type_to_paths, meta_provider


def _infer_content_types_from_paths(
    paths: List[str],
    content_type_provider: Callable[[str], ContentType],
) -> Dict[ContentType, List[str]]:
    content_type_to_paths = defaultdict(list)
    for path in paths:
        if not path.endswith("/"):
            content_type_to_paths[content_type_provider(path)].append(path)
    return content_type_to_paths


def _expand_prefix_paths(
    urls: List[S3Url],
    content_type_provider: Callable[[str], ContentType],
    **s3_client_kwargs,
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    assert len(urls) == 1, f"Expected 1 S3 prefix, found {len(urls)}."
    objects = list(
        filter_objects_by_prefix(urls[0].bucket, urls[0].key, **s3_client_kwargs)
    )
    paths = list(
        objects_to_paths(
            urls[0].bucket,
            objects,
        )
    )
    meta_cache: Dict[str, BlockMetadata] = {
        path: BlockMetadata(
            num_rows=None,
            size_bytes=objects[i]["ContentLength"],
            schema=None,
            input_files=[],
            exec_stats=None,
        )
        for i, path in enumerate(paths)
    }
    content_type_to_paths = _infer_content_types_from_paths(
        paths,
        content_type_provider,
    )
    return content_type_to_paths, CachedFileMetadataProvider(meta_cache)


def _expand_paths_by_content_type(
    base_paths: Union[str, List[str]],
    base_urls: List[S3Url],
    content_type_provider: Callable[[str], ContentType],
    path_type: S3PathType,
    user_fs: Optional[Union[S3FileSystem, s3fs.S3FileSystem]],
    resolved_fs: S3FileSystem,
    **s3_client_kwargs,
) -> Tuple[Dict[ContentType, List[str]], CachedFileMetadataProvider]:
    if path_type == S3PathType.MANIFEST:
        content_type_to_paths, meta_provider = _expand_manifest_paths(
            base_paths,
            resolved_fs,
            content_type_provider,
        )
    elif path_type == S3PathType.PREFIX:
        content_type_to_paths, meta_provider = _expand_prefix_paths(
            base_urls,
            content_type_provider,
            **s3_client_kwargs,
        )
    elif path_type == S3PathType.FILES_AND_FOLDERS:
        # TODO(pdames): Only allow files and call get_object(file_path)?
        base_paths, file_infos = DefaultFileMetadataProvider().expand_paths(
            base_paths, resolved_fs
        )
        file_sizes = [file_info.size for file_info in file_infos]
        meta_provider = CachedFileMetadataProvider(
            {
                path: BlockMetadata(
                    num_rows=None,
                    size_bytes=file_sizes[i],
                    schema=None,
                    input_files=[],
                    exec_stats=None,
                )
                for i, path in enumerate(base_paths)
            }
        )
        content_type_to_paths = _infer_content_types_from_paths(
            base_paths,
            content_type_provider,
        )
    else:
        raise NotImplementedError(f"Unsupported S3 path type: {path_type}")
    # TODO(pdames): normalize S3 file paths before adding them to either
    #  content_type_to_paths or meta_provider
    # normalize S3 file paths for each content type based on the filesystem
    for content_type, paths in content_type_to_paths.items():
        paths, urls = _normalize_s3_paths_for_filesystem(
            paths,
            user_fs,
        )
        content_type_to_paths[content_type] = paths
    # normalize block metadata provider S3 file paths based on the filesystem
    meta_provider = CachedFileMetadataProvider(
        {
            _normalize_s3_paths_for_filesystem(path, user_fs)[0][0]: metadata
            for path, metadata in meta_provider.get_meta_cache().items()
        }
    )
    return content_type_to_paths, meta_provider


class RedshiftDatasource(Datasource[Union[ArrowRow, Any]]):
    def prepare_read(
        self,
        parallelism: int,
        paths: Union[str, List[str]],
        content_type_provider: Callable[[str], ContentType],
        path_type: S3PathType = S3PathType.MANIFEST,
        filesystem: Optional[Union[S3FileSystem, s3fs.S3FileSystem]] = None,
        columns: Optional[List[str]] = None,
        schema: Optional[pa.Schema] = None,
        unload_args: RedshiftUnloadTextArgs = RedshiftUnloadTextArgs(),
        partitioning: HivePartitionParser = None,
        open_stream_args: Optional[Dict[str, Any]] = None,
        read_kwargs_provider: Optional[ReadKwargsProvider] = None,
        **s3_client_kwargs,
    ) -> List[ReadTask]:
        # default to pyarrow.fs.S3FileSystem if no filesystem given
        if filesystem is None:
            filesystem = S3FileSystem()
        # normalize s3 paths to work with the filesystem provided
        paths, urls = _normalize_s3_paths_for_filesystem(paths, filesystem)
        paths, resolved_fs = _resolve_paths_and_filesystem(
            paths,
            filesystem,
        )
        # find all files in manifests, prefixes, and folders
        content_type_to_paths, meta_provider = _expand_paths_by_content_type(
            paths,
            urls,
            content_type_provider,
            path_type,
            filesystem,
            resolved_fs,
            **s3_client_kwargs,
        )
        num_content_types = len(content_type_to_paths)
        if num_content_types > 1 and not schema:
            # infer schema from a single parquet file
            # TODO (pdames): read verbose manifest schema if available, and infer
            #  schema from a sample parquet dataset if not
            path = content_type_to_paths[ContentType.PARQUET][0]
            with resolved_fs.open_input_file(path, **open_stream_args) as f:
                schema = pq.read_schema(f)
        content_type_to_reader = {
            ContentType.PARQUET: ParquetBaseDatasource(),
            ContentType.CSV: CSVDatasource(),
        }
        all_read_tasks = []
        for content_type, paths in content_type_to_paths.items():
            reader = content_type_to_reader.get(content_type)
            assert reader, f"No datasource found for: {content_type}"
            prepare_read_kwargs = {
                "parallelism": parallelism,
                "paths": paths,
                "filesystem": resolved_fs,
                "schema": schema,
                "meta_provider": meta_provider,
                "partitioning": partitioning,
            }
            if content_type == ContentType.PARQUET:
                if columns:
                    prepare_read_kwargs["columns"] = columns
            elif content_type in DELIMITED_TEXT_CONTENT_TYPES:
                prepare_read_kwargs.update(
                    unload_args.to_arrow_reader_kwargs(columns, schema)
                )
            else:
                raise NotImplementedError(f"Unsupported content type: {content_type}")
            # merge any provided reader kwargs for this content type with those
            # inferred from Redshift UNLOAD args
            if read_kwargs_provider:
                prepare_read_kwargs = read_kwargs_provider(
                    content_type,
                    prepare_read_kwargs,
                )
            # explicitly specified `open_stream_args` override those inferred
            # from Redshift UNLOAD args
            if open_stream_args:
                prepare_read_kwargs["open_stream_args"] = open_stream_args
            read_tasks = reader.prepare_read(**prepare_read_kwargs)
            all_read_tasks.extend(read_tasks)
        return all_read_tasks

    def do_write(
        self,
        blocks: List[ObjectRef[Block]],
        metadata: List[BlockMetadata],
        path: str,
        dataset_uuid: str,
        filesystem: Optional[FileSystem] = None,
        try_create_dir: bool = True,
        open_stream_args: Optional[Dict[str, Any]] = None,
        block_path_provider: BlockWritePathProvider = DefaultBlockWritePathProvider(),
        write_args_fn: Callable[[], Dict[str, Any]] = lambda: {},
        _block_udf: Optional[Callable[[Block], Block]] = None,
        **write_args,
    ) -> List[ObjectRef[WriteResult]]:
        if filesystem is None:
            filesystem = S3FileSystem()
        paths, _ = _normalize_s3_paths_for_filesystem(path, filesystem)
        paths, filesystem = _resolve_paths_and_filesystem(paths, filesystem)
        assert len(paths) == 1, f"Expected 1 write path, found {len(paths)}."
        path = paths[0]
        block_path_provider = CapturingBlockWritePathProvider(block_path_provider)
        writer = ParquetBaseDatasource()
        write_results = writer.do_write(
            blocks,
            metadata,
            path,
            dataset_uuid,
            filesystem,
            try_create_dir,
            open_stream_args,
            block_path_provider,
            write_args_fn,
            _block_udf,
            **write_args,
        )
        # append a summary of this write operation in the last write result
        rwr = RedshiftWriteResult()
        rwr.metadata = metadata
        rwr.path = path
        rwr.dataset_uuid = dataset_uuid
        rwr.block_write_path_provider = block_path_provider
        rwr.content_type = ContentType.PARQUET.value
        rwr.content_encoding = ContentEncoding.IDENTITY.value
        rwr.filesystem = filesystem
        rwr_obj_ref = ray.put(rwr)
        write_results.append(rwr_obj_ref)
        return write_results

    def on_write_complete(self, write_results: List[WriteResult], **kwargs) -> None:
        # TODO (pdames): time latency of this operation - overall redshift write times
        #  are 2-3x pure read_parquet_fast() times
        # restore the write operation summary from the last write result
        result: RedshiftWriteResult = write_results[len(write_results) - 1]
        write_path_args = result.block_write_path_provider.write_path_kwargs
        blocks_written = len(write_path_args)
        expected_blocks_written = len(result.metadata)
        # TODO(pdames): Corner cases where mismatch is expected? Emply blocks?
        #  Blocks filtered/split/merged to more/less write paths?
        assert blocks_written == expected_blocks_written, (
            f"Dataset write result validation failed. Found "
            f"{blocks_written}/{expected_blocks_written} Dataset blocks "
            f"written. Refusing to commit Redshift Manifest."
        )
        manifest_entries = ManifestEntryList()
        for block_idx, path in enumerate(write_path_args.keys()):
            file_info = result.filesystem.get_file_info(path)
            if file_info.type == FileType.File:
                content_length = file_info.size
            else:
                raise FileNotFoundError(ENOENT, strerror(ENOENT), path)
            num_rows = result.metadata[block_idx].num_rows
            source_content_length = result.metadata[block_idx].size_bytes
            manifest_entry_meta = ManifestMeta.of(
                int(num_rows) if num_rows is not None else None,
                int(content_length) if content_length is not None else None,
                result.content_type,
                result.content_encoding,
                int(source_content_length) if source_content_length else None,
            )
            parsed_url = parse_s3_url(path)
            manifest_entry = ManifestEntry.of(
                parsed_url.url,
                manifest_entry_meta,
            )
            manifest_entries.append(manifest_entry)
        manifest = Manifest.of(manifest_entries)
        manifest_path = f"{result.path}/manifest"
        logger.debug(f"Write succeeded for Dataset ID: {result.dataset_uuid}")
        with result.filesystem.open_output_stream(
            manifest_path,
            # Also See:
            # docs.aws.amazon.com/AmazonS3/latest/API/RESTCommonRequestHeaders.html
            # Arrow s3fs.cc: tinyurl.com/2axa6m9m
            metadata={"Content-Type": ContentType.JSON.value},
        ) as f:
            f.write(json.dumps(manifest).encode("utf-8"))
        logger.debug(f"Manifest committed to: {manifest_path}")



from typing import Iterable, List, Optional, Set, Tuple, Union

DeltaPosition = Optional[int]
NumericDeltaPosition = Union[int, float]  # float is added here to support math.inf
DeltaRange = Tuple[DeltaPosition, DeltaPosition]


def merge_intervals(intervals: Set[DeltaRange]) -> Set[DeltaRange]:
    """Merges a set of N input intervals into a minimal number of output intervals.

    All input intervals will be merged into a minimal number of output intervals in O(N log N) time.

    Example:
        >>> merge_intervals((3, 9), (8, 12), (15, 19))
        ((3, 12), (15, 19))

    Example:
        >>> merge_intervals((3, 9), (None, 15), (13, 30))
        ((None, 30))

    Args:
        intervals: A list of intervals with an int type representing finite, closed bounded values and
            a None type representing infinity.

    Returns:
        A minimal number of output intervals
    """
    merged: Set[DeltaRange] = set()
    intervals_list: List[DeltaRange] = list(intervals)

    if len(intervals_list) == 0:
        return merged

    _to_numeric_values(intervals_list)
    intervals_list.sort()  # sort by starting range numbers

    merge_start, merge_end = None, None
    for interval in intervals_list:
        start, end = interval
        if start > end:
            raise ValueError(
                f"Invalid stream position range interval: ({start}, {end})"
            )

        if merge_start is None and merge_end is None:
            merge_start, merge_end = start, end
            continue

        if merge_end < start:
            # add current merge interval if no overlap, begin new interval
            _add_merged_interval(merged, merge_start, merge_end)
            merge_start, merge_end = start, end
        elif merge_end < end:
            merge_end = end  # expand current merge interval if there is an overlap

    # add final merge interval
    _add_merged_interval(merged, merge_start, merge_end)

    return merged


def _add_merged_interval(
    result_set: set, start: NumericDeltaPosition, end: NumericDeltaPosition
):
    start_pos: DeltaPosition = start if isinstance(start, int) else None
    end_pos: DeltaPosition = end if isinstance(end, int) else None
    result_set.add((start_pos, end_pos))


def _to_numeric_values(intervals_list: List[DeltaRange]):
    for i, interval in enumerate(intervals_list):
        start, end = _get_validated_interval(interval)
        if start is None:
            start = float("-inf")
        if end is None:
            end = float("inf")

        intervals_list[i] = (start, end)


def _get_validated_interval(interval: DeltaRange) -> DeltaRange:
    if not isinstance(interval, Iterable) or len(interval) != 2:
        raise ValueError(f"Interval {interval} must be a tuple of size 2")

    start, end = interval
    if not (isinstance(start, int) or start is None) or not (
        isinstance(end, int) or end is None
    ):
        raise ValueError(
            f"Invalid stream position value types: "
            f"({start}, {end}) - ({type(start), type(end)})"
        )

    return start, end



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List

import pyarrow as pa

from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.storage import DeltaLocator


class ManifestEntryStats(dict):
    """Holds computed statistics for one or more manifest entries (tables) and their corresponding delta locator.

    To be stored/retrieved from a file system (ex: S3).
    """

    @staticmethod
    def of(
        manifest_entries_stats: List[StatsResult], delta_locator: DeltaLocator
    ) -> ManifestEntryStats:
        """
        Creates a stats container that represents a particular manifest.

        `manifest_entries_stats` are a list of distinct stats for each manifest entry file
        tied to this manifest. `delta_locator` is provided as a reference to the delta where the
        manifest entries reside.
        """

        mes = ManifestEntryStats()
        mes["deltaLocator"] = delta_locator
        mes["stats"] = manifest_entries_stats
        mes["pyarrowVersion"] = pa.__version__
        return mes

    @staticmethod
    def build_from_dict(manifest_entries_stats: dict) -> ManifestEntryStats:
        stats_res_list = []
        for stats_res in manifest_entries_stats["stats"]:
            stats_res_list.append(
                StatsResult.of(stats_res["rowCount"], stats_res["pyarrowTableBytes"])
            )
        return ManifestEntryStats.of(
            stats_res_list, manifest_entries_stats["deltaLocator"]
        )

    @property
    def delta_locator(self) -> DeltaLocator:
        """Reference to the delta that holds the manifest entries

        Returns:
            A delta locator object
        """
        val: Dict[str, Any] = self.get("deltaLocator")
        if val is not None and not isinstance(val, DeltaLocator):
            self["deltaLocator"] = val = DeltaLocator(val)
        return val

    @property
    def stats(self) -> List[StatsResult]:
        """
        Returns a list of distinct stats for each manifest entry file.
        """
        val = self["stats"]
        return [StatsResult(_) for _ in val] if val else []

    @property
    def pyarrow_version(self) -> str:
        """
        Read-only property which returns the PyArrow version number as it was written into a file system.
        """
        return self.get("pyarrowVersion")



from deltacat.utils.common import env_string

# Environment variables
DELTACAT_LOG_LEVEL = env_string("DELTACAT_LOG_LEVEL", "DEBUG")
APPLICATION_LOG_LEVEL = env_string("APPLICATION_LOG_LEVEL", "DEBUG")

# Byte Units
BYTES_PER_KIBIBYTE = 2**10
BYTES_PER_MEBIBYTE = 2**20
BYTES_PER_GIBIBYTE = 2**30
BYTES_PER_TEBIBYTE = 2**40
BYTES_PER_PEBIBYTE = 2**50

# Inflation multiplier from snappy-compressed parquet to pyarrow.
# This should be kept larger than actual average inflation multipliers.
# Note that this is a very rough guess since actual observed pyarrow
# inflation multiplier for snappy-compressed parquet is about 5.45X for
# all rows, but here we're trying to guess the inflation multipler for just
# a primary key SHA1 digest and sort key columns (which could be all columns
# of the table in the worst case, but here we're assuming that they
# represent no more than ~1/4th of the total table bytes)
PYARROW_INFLATION_MULTIPLIER = 2.5

# Inflation multiplier from snappy-compressed parquet to pyarrow for all columns.
PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS = 6

PRIMARY_KEY_INDEX_WRITE_BOTO3_CONFIG = {
    "retries": {"max_attempts": 25, "mode": "standard"}
}



import csv
import io
import logging
import math
from typing import Any, Callable, Dict, Iterable, List, Optional

import pandas as pd
import pyarrow as pa
from fsspec import AbstractFileSystem
from ray.data.datasource import BlockWritePathProvider

from deltacat import logs
from deltacat.types.media import (
    DELIMITED_TEXT_CONTENT_TYPES,
    EXPLICIT_COMPRESSION_CONTENT_TYPES,
    TABULAR_CONTENT_TYPES,
    ContentEncoding,
    ContentType,
)
from deltacat.utils.common import ContentTypeKwargsProvider, ReadKwargsProvider
from deltacat.utils.performance import timed_invocation

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


CONTENT_TYPE_TO_PD_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pd.read_csv,
    ContentType.TSV.value: pd.read_csv,
    ContentType.CSV.value: pd.read_csv,
    ContentType.PSV.value: pd.read_csv,
    ContentType.PARQUET.value: pd.read_parquet,
    ContentType.FEATHER.value: pd.read_feather,
    ContentType.ORC.value: pd.read_orc,
    ContentType.JSON.value: pd.read_json,
}


class ReadKwargsProviderPandasCsvPureUtf8(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""

    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            include_columns = (
                self.include_columns if self.include_columns else kwargs.get("usecols")
            )
            if not include_columns:
                # read all columns as strings
                kwargs["dtype"] = str
            else:
                # read only the included columns as strings
                kwargs["dtype"] = {column_name: str for column_name in include_columns}
            # use the fastest available engine for pure utf-8 reads
            kwargs["engine"] = "pyarrow"
        return kwargs


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "sep": "\t",
            "header": None,
            "na_values": [""],
            "keep_default_na": False,
        }
    if content_type == ContentType.TSV.value:
        return {"sep": "\t", "header": None}
    if content_type == ContentType.CSV.value:
        return {"sep": ",", "header": None}
    if content_type == ContentType.PSV.value:
        return {"sep": "|", "header": None}
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.ORC.value,
        ContentType.JSON.value,
    }:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


ENCODING_TO_PD_COMPRESSION: Dict[str, str] = {
    ContentEncoding.GZIP.value: "gzip",
    ContentEncoding.BZIP2.value: "bz2",
    ContentEncoding.IDENTITY.value: "none",
}


def slice_dataframe(
    dataframe: pd.DataFrame, max_len: Optional[int]
) -> List[pd.DataFrame]:
    """
    Iteratively create dataframe slices.
    """
    if max_len is None:
        return [dataframe]
    dataframes = []
    num_slices = math.ceil(len(dataframe) / max_len)
    for i in range(num_slices):
        dataframes.append(dataframe[i * max_len : (i + 1) * max_len])
    return dataframes


def concat_dataframes(dataframes: List[pd.DataFrame]) -> Optional[pd.DataFrame]:
    if dataframes is None or not len(dataframes):
        return None
    if len(dataframes) == 1:
        return next(iter(dataframes))
    return pd.concat(dataframes, axis=0, copy=False)


def _add_column_kwargs(
    content_type: str,
    column_names: Optional[List[str]],
    include_columns: Optional[List[str]],
    kwargs: Dict[str, Any],
):

    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        kwargs["names"] = column_names
        kwargs["usecols"] = include_columns
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            kwargs["columns"]: include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}"
                )


def s3_file_to_dataframe(
    s3_url: str,
    content_type: str,
    content_encoding: str,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pd_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    **s3_client_kwargs,
) -> pd.DataFrame:

    from deltacat.aws import s3u as s3_utils

    logger.debug(
        f"Reading {s3_url} to Pandas. Content type: {content_type}. "
        f"Encoding: {content_encoding}"
    )
    s3_obj = s3_utils.get_object_at_url(s3_url, **s3_client_kwargs)
    logger.debug(f"Read S3 object from {s3_url}: {s3_obj}")
    pd_read_func = CONTENT_TYPE_TO_PD_READ_FUNC[content_type]
    args = [io.BytesIO(s3_obj["Body"].read())]
    kwargs = content_type_to_reader_kwargs(content_type)
    _add_column_kwargs(content_type, column_names, include_columns, kwargs)

    if content_type in EXPLICIT_COMPRESSION_CONTENT_TYPES:
        kwargs["compression"] = ENCODING_TO_PD_COMPRESSION.get(
            content_encoding, "infer"
        )
    if pd_read_func_kwargs_provider:
        kwargs = pd_read_func_kwargs_provider(content_type, kwargs)
    logger.debug(f"Reading {s3_url} via {pd_read_func} with kwargs: {kwargs}")
    dataframe, latency = timed_invocation(pd_read_func, *args, **kwargs)
    logger.debug(f"Time to read {s3_url} into Pandas Dataframe: {latency}s")
    return dataframe


def dataframe_size(dataframe: pd.DataFrame) -> int:
    # TODO (pdames): inspect latency vs. deep memory usage inspection
    return int(dataframe.memory_usage().sum())


def write_csv(
    dataframe: pd.DataFrame, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:
    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            dataframe.to_csv(out, **kwargs)


def write_parquet(
    dataframe: pd.DataFrame, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:
    with filesystem.open(path, "wb") as f:
        dataframe.to_parquet(f, **kwargs)


def write_feather(
    dataframe: pd.DataFrame, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:
    with filesystem.open(path, "wb") as f:
        dataframe.to_feather(f, **kwargs)


def write_json(
    dataframe: pd.DataFrame, path: str, *, filesystem: AbstractFileSystem, **kwargs
) -> None:
    with filesystem.open(path, "wb") as f:
        # TODO (pdames): Add support for client-specified compression types.
        with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
            dataframe.to_json(out, **kwargs)


CONTENT_TYPE_TO_PD_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV: write_csv,
    ContentType.TSV.value: write_csv,
    ContentType.CSV.value: write_csv,
    ContentType.PSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
    ContentType.FEATHER.value: write_feather,
    ContentType.JSON.value: write_json,
}


def content_type_to_writer_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "sep": "\t",
            "header": False,
            "na_rep": [""],
            "line_terminator": "\n",
            "quoting": csv.QUOTE_NONE,
            "index": False,
        }
    if content_type == ContentType.TSV.value:
        return {
            "sep": "\t",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.CSV.value:
        return {
            "sep": ",",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.PSV.value:
        return {
            "sep": "|",
            "header": False,
            "line_terminator": "\n",
            "index": False,
        }
    if content_type == ContentType.PARQUET.value:
        return {"index": False}
    if content_type == ContentType.FEATHER.value:
        return {}
    if content_type == ContentType.JSON.value:
        return {"index": False}
    raise ValueError(f"Unsupported content type: {content_type}")


def dataframe_to_file(
    dataframe: pd.DataFrame,
    base_path: str,
    file_system: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    content_type: str = ContentType.PARQUET.value,
    **kwargs,
) -> None:
    """
    Writes the given Pandas Dataframe to a file.
    """
    writer = CONTENT_TYPE_TO_PD_WRITE_FUNC.get(content_type)
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    if not writer:
        raise NotImplementedError(
            f"Pandas writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PD_WRITE_FUNC.keys}"
        )
    path = block_path_provider(base_path)
    writer(dataframe, path, filesystem=file_system, **writer_kwargs)



import logging
from typing import List, Tuple

import numpy as np
import pyarrow as pa
import ray
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import PrimaryKeyIndexVersionLocator
from deltacat.compute.compactor.utils import primary_key_index as pki

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def group_file_records_by_pk_hash_bucket(
    pki_table: pa.Table, num_buckets: int
) -> np.ndarray:
    # generate the new table for each new hash bucket
    hash_bucket_to_indices = pki.group_record_indices_by_hash_bucket(
        pki_table,
        num_buckets,
    )
    hash_bucket_to_table = np.empty([num_buckets], dtype="object")
    for hash_bucket, indices in enumerate(hash_bucket_to_indices):
        if indices:
            hash_bucket_to_table[hash_bucket] = pki_table.take(indices)
    return hash_bucket_to_table


@ray.remote(num_cpus=1, num_returns=2)
def rehash_bucket(
    hash_bucket_index: int,
    s3_bucket: str,
    old_pki_version_locator: PrimaryKeyIndexVersionLocator,
    num_buckets: int,
    num_groups: int,
) -> Tuple[np.ndarray, List[ObjectRef]]:

    logger.info(f"Starting rehash bucket task...")
    tables = pki.download_hash_bucket_entries(
        s3_bucket,
        hash_bucket_index,
        old_pki_version_locator,
    )
    prior_pk_index_table = pa.concat_tables(tables)
    hash_bucket_to_table = group_file_records_by_pk_hash_bucket(
        prior_pk_index_table,
        num_buckets,
    )
    hash_bucket_group_to_obj_id, object_refs = pki.group_hash_bucket_indices(
        hash_bucket_to_table,
        num_buckets,
        num_groups,
    )
    logger.info(f"Finished rehash bucket task...")
    return hash_bucket_group_to_obj_id, object_refs



from enum import Enum
from typing import Any, List, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from ray.data._internal.arrow_block import ArrowRow
from ray.data.dataset import Dataset

LocalTable = Union[pa.Table, pd.DataFrame, np.ndarray]
LocalDataset = List[LocalTable]
DistributedDataset = Dataset[Union[ArrowRow, np.ndarray, Any]]


class DeltaType(str, Enum):
    APPEND = "append"
    UPSERT = "upsert"
    DELETE = "delete"


class LifecycleState(str, Enum):
    UNRELEASED = "unreleased"
    ACTIVE = "active"
    DEPRECATED = "deprecated"
    BETA = "beta"
    DELETED = "deleted"


class CommitState(str, Enum):
    STAGED = "staged"
    COMMITTED = "committed"
    DEPRECATED = "deprecated"


class SchemaConsistencyType(str, Enum):
    """
    Schemas are optional for DeltaCAT tables and can be used to inform the data
    consistency checks run for each field. If a schema is present, it can be
    used to enforce the following column-level data consistency policies at
    table load time:

    NONE: No consistency checks are run. May be mixed with the below two
    policies by specifying column names to pass through together with
    column names to coerce/validate.

    COERCE: Coerce fields to fit the schema whenever possible. An explicit
    subset of column names to coerce may optionally be specified.

    VALIDATE: Raise an error for any fields that don't fit the schema. An
    explicit subset of column names to validate may optionally be specified.
    """

    NONE = "none"
    COERCE = "coerce"
    VALIDATE = "validate"



import logging
from itertools import chain
from typing import Generator, List, Optional, Tuple

import numpy as np
import pyarrow as pa
import ray
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import DeltaAnnotated, DeltaFileEnvelope, SortKey
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelopeGroups
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.compute.compactor.utils.primary_key_index import (
    group_hash_bucket_indices,
    group_record_indices_by_hash_bucket,
)
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.types.media import StorageType
from deltacat.utils.common import sha1_digest

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

_PK_BYTES_DELIMITER = b"L6kl7u5f"

HashBucketGroupToObjectId = np.ndarray
HashBucketResult = Tuple[
    HashBucketGroupToObjectId, List[ObjectRef[DeltaFileEnvelopeGroups]]
]


def _group_by_pk_hash_bucket(
    table: pa.Table, num_buckets: int, primary_keys: List[str]
) -> np.ndarray:

    # generate the primary key digest column
    all_pk_column_fields = []
    for pk_name in primary_keys:
        # casting a primary key column to numpy also ensures no nulls exist
        column_fields = table[pk_name].to_numpy()
        all_pk_column_fields.append(column_fields)
    hash_column_generator = _hash_pk_bytes_generator(all_pk_column_fields)
    table = sc.append_pk_hash_column(table, hash_column_generator)

    # drop primary key columns to free up memory
    table = table.drop(primary_keys)

    # group hash bucket record indices
    hash_bucket_to_indices = group_record_indices_by_hash_bucket(
        table,
        num_buckets,
    )

    # generate the ordered record number column
    hash_bucket_to_table = np.empty([num_buckets], dtype="object")
    for hb, indices in enumerate(hash_bucket_to_indices):
        if indices:
            hash_bucket_to_table[hb] = sc.append_record_idx_col(
                table.take(indices),
                indices,
            )
    return hash_bucket_to_table


def _hash_pk_bytes_generator(all_column_fields) -> Generator[bytes, None, None]:
    for field_index in range(len(all_column_fields[0])):
        bytes_to_join = []
        for column_fields in all_column_fields:
            bytes_to_join.append(bytes(str(column_fields[field_index]), "utf-8"))
        yield sha1_digest(_PK_BYTES_DELIMITER.join(bytes_to_join))


def _group_file_records_by_pk_hash_bucket(
    annotated_delta: DeltaAnnotated,
    num_hash_buckets: int,
    primary_keys: List[str],
    sort_key_names: List[str],
    deltacat_storage=unimplemented_deltacat_storage,
) -> Optional[DeltaFileEnvelopeGroups]:

    # read input parquet s3 objects into a list of delta file envelopes
    delta_file_envelopes = _read_delta_file_envelopes(
        annotated_delta,
        primary_keys,
        sort_key_names,
        deltacat_storage,
    )
    if delta_file_envelopes is None:
        return None

    # group the data by primary key hash value
    hb_to_delta_file_envelopes = np.empty([num_hash_buckets], dtype="object")
    for dfe in delta_file_envelopes:
        hash_bucket_to_table = _group_by_pk_hash_bucket(
            dfe.table,
            num_hash_buckets,
            primary_keys,
        )
        for hb, table in enumerate(hash_bucket_to_table):
            if table:
                if hb_to_delta_file_envelopes[hb] is None:
                    hb_to_delta_file_envelopes[hb] = []
                hb_to_delta_file_envelopes[hb].append(
                    DeltaFileEnvelope.of(
                        dfe.stream_position, dfe.file_index, dfe.delta_type, table
                    )
                )
    return hb_to_delta_file_envelopes


def _read_delta_file_envelopes(
    annotated_delta: DeltaAnnotated,
    primary_keys: List[str],
    sort_key_names: List[str],
    deltacat_storage=unimplemented_deltacat_storage,
) -> Optional[List[DeltaFileEnvelope]]:

    columns_to_read = list(chain(primary_keys, sort_key_names))
    tables = deltacat_storage.download_delta(
        annotated_delta,
        max_parallelism=1,
        columns=columns_to_read,
        storage_type=StorageType.LOCAL,
    )
    annotations = annotated_delta.annotations
    assert (
        len(tables) == len(annotations),
        f"Unexpected Error: Length of downloaded delta manifest tables "
        f"({len(tables)}) doesn't match the length of delta manifest "
        f"annotations ({len(annotations)}).",
    )
    if not tables:
        return None

    delta_file_envelopes = []
    for i, table in enumerate(tables):
        delta_file = DeltaFileEnvelope.of(
            annotations[i].annotation_stream_position,
            annotations[i].annotation_file_index,
            annotations[i].annotation_delta_type,
            table,
        )
        delta_file_envelopes.append(delta_file)
    return delta_file_envelopes


@ray.remote(num_returns=2)
def hash_bucket(
    annotated_delta: DeltaAnnotated,
    primary_keys: List[str],
    sort_keys: List[SortKey],
    num_buckets: int,
    num_groups: int,
    deltacat_storage=unimplemented_deltacat_storage,
) -> HashBucketResult:

    logger.info(f"Starting hash bucket task...")
    sort_key_names = [key.key_name for key in sort_keys]
    delta_file_envelope_groups = _group_file_records_by_pk_hash_bucket(
        annotated_delta,
        num_buckets,
        primary_keys,
        sort_key_names,
        deltacat_storage,
    )
    hash_bucket_group_to_obj_id, object_refs = group_hash_bucket_indices(
        delta_file_envelope_groups,
        num_buckets,
        num_groups,
    )
    logger.info(f"Finished hash bucket task...")
    return hash_bucket_group_to_obj_id, object_refs



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import numpy as np

from deltacat.storage import Locator


class DeltaFileLocator(Locator, tuple):
    @staticmethod
    def of(
        is_src_delta: np.bool_, stream_position: np.int64, file_index: np.int32
    ) -> DeltaFileLocator:
        """
        Create a Delta File Locator tuple that can be used to uniquely identify
        and retrieve a file from any compaction job run input Delta.

        Args:
            is_src_delta: True if this Delta File Locator is
                pointing to a file from the uncompacted source table, False if
                this Locator is pointing to a file in the compacted destination
                table.

            stream_position: Stream position of the Delta that
                contains this file.

            file_index: Index of the file in the Delta Manifest.

        Returns:
            delta_file_locator: The Delta File Locator Tuple as
            (is_source_delta, stream_position, file_index).
        """
        return DeltaFileLocator(
            (
                is_src_delta,
                stream_position,
                file_index,
            )
        )

    @property
    def is_source_delta(self) -> np.bool_:
        return self[0]

    @property
    def stream_position(self) -> np.int64:
        return self[1]

    @property
    def file_index(self) -> np.int32:
        return self[2]

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        return f"{self[0]}|{self[1]}|{self[2]}"



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Dict

from deltacat.compute.stats.models.delta_stats import DeltaStats


class PartitionStats(dict):
    @staticmethod
    def of(
        delta_stats: Dict[DeltaStats], partition_canonical_string: str
    ) -> PartitionStats:
        ps = PartitionStats()
        ps["delta_stats"] = delta_stats
        ps["partition_canonical_string"] = partition_canonical_string
        return ps

    @staticmethod
    def build_from_dict(partition_stats: str) -> PartitionStats:
        delta_stats_dict = {}
        for stream_position, delta_stats in partition_stats["delta_stats"].items():
            delta_stats_dict[stream_position] = DeltaStats.build_from_dict(delta_stats)
        return PartitionStats.of(
            delta_stats_dict, partition_stats["partition_canonical_string"]
        )

    @property
    def delta_stats(self) -> Dict[DeltaStats]:
        return self["delta_stats"]

    @property
    def partition_canonical_string(self) -> str:
        return self["partition_canonical_string"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import logging
from enum import Enum
from typing import List

import pyarrow as pa

from deltacat import logs
from deltacat.storage import PartitionLocator

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

MAX_SORT_KEYS_BIT_WIDTH: int = 256


class SortOrder(str, Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class SortKey(tuple):
    @staticmethod
    def of(key_name: str, sort_order: SortOrder = SortOrder.ASCENDING) -> SortKey:
        """
        Create a sort key from a field name to use as the sort key, and
        the sort order for this key. If no sort order is specified, then the
        data will be sorted in ascending order by default. Note that compaction
        always keeps the LAST occurrence of this key post-sort. For example, if
        you used an integer column as your sort key which contained the values
        [2, 1, 3] specifying SortOrder.ASCENDING would ensure that the
        value [3] is kept over [2, 1], and specifying SortOrder.DESCENDING
        would ensure that [1] is kept over [2, 3].
        """
        return SortKey((key_name, sort_order.value))

    @staticmethod
    def validate_sort_keys(
        source_partition_locator: PartitionLocator,
        sort_keys: List[SortKey],
        deltacat_storage,
    ) -> int:
        """
        Validates the input sort keys to ensure that they are unique, are using
        a valid sort key model, are all fixed-width data types, and that the
        sum of bit widths across sort key data types is less-than-or-equal-to
        256. Returns the sum of bit widths across all sort keys.
        """
        total_sort_keys_bit_width = 0
        if sort_keys:
            sort_key_names = [key.key_name for key in sort_keys]
            assert len(sort_key_names) == len(
                set(sort_key_names)
            ), f"Sort key names must be unique: {sort_key_names}"
            stream_locator = source_partition_locator.stream_locator
            table_version_schema = deltacat_storage.get_table_version_schema(
                stream_locator.namespace,
                stream_locator.table_name,
                stream_locator.table_version,
            )
            if isinstance(table_version_schema, pa.Schema):
                for sort_key_name in sort_key_names:
                    pa_field: pa.Field = pa.Schema.field(sort_key_name)
                    pa_type: pa.DataType = pa_field.type
                    try:
                        total_sort_keys_bit_width += pa_type.bit_width
                        if total_sort_keys_bit_width > MAX_SORT_KEYS_BIT_WIDTH:
                            raise ValueError(
                                f"Total length of sort keys "
                                f"({total_sort_keys_bit_width}) is greater "
                                f"than the max supported bit width for all "
                                f"sort keys ({MAX_SORT_KEYS_BIT_WIDTH})"
                            )
                    except ValueError as e:
                        raise ValueError(
                            f"Unable to get bit width of sort key: {pa_field}. "
                            f"Please ensure that all sort keys are fixed-size "
                            f"PyArrow data types."
                        ) from e
            else:
                logger.warning(
                    f"Unable to estimate sort key bit width for schema type "
                    f"{type(table_version_schema)}. This compaction job run "
                    f"may run out of memory, run more slowly, or underutilize "
                    f"available resources. To fix this, either remove the "
                    f"sort keys or provide a PyArrow schema."
                )
                total_sort_keys_bit_width = MAX_SORT_KEYS_BIT_WIDTH
        return total_sort_keys_bit_width

    @property
    def key_name(self) -> str:
        return self[0]

    @property
    def sort_order(self) -> SortOrder:
        return SortOrder(self[1])



import logging
import math
from typing import Dict, List, Optional, Tuple

from deltacat import logs
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.constants import BYTES_PER_MEBIBYTE, PYARROW_INFLATION_MULTIPLIER
from deltacat.storage import Delta, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def discover_deltas(
    source_partition_locator: PartitionLocator,
    start_position_exclusive: Optional[int],
    end_position_inclusive: int,
    deltacat_storage=unimplemented_deltacat_storage,
) -> List[Delta]:

    stream_locator = source_partition_locator.stream_locator
    namespace = stream_locator.namespace
    table_name = stream_locator.table_name
    table_version = stream_locator.table_version
    partition_values = source_partition_locator.partition_values
    deltas_list_result = deltacat_storage.list_deltas(
        namespace,
        table_name,
        partition_values,
        table_version,
        start_position_exclusive,
        end_position_inclusive,
        True,
    )
    deltas = deltas_list_result.all_items()
    if not deltas:
        raise RuntimeError(
            f"Unexpected Error: Couldn't find any deltas to "
            f"compact in delta stream position range "
            f"('{start_position_exclusive}', "
            f"'{end_position_inclusive}']. Source partition: "
            f"{source_partition_locator}"
        )
    if start_position_exclusive:
        first_delta = deltas.pop(0)
        logger.info(
            f"Removed exclusive start delta w/ expected stream "
            f"position '{start_position_exclusive}' from deltas to "
            f"compact: {first_delta}"
        )
    logger.info(
        f"Count of deltas to compact in delta stream "
        f"position range ('{start_position_exclusive}', "
        f"'{end_position_inclusive}']: {len(deltas)}. Source "
        f"partition: '{source_partition_locator}'"
    )
    return deltas


def limit_input_deltas(
    input_deltas: List[Delta],
    cluster_resources: Dict[str, float],
    hash_bucket_count: int,
    min_pk_index_pa_bytes: int,
    user_hash_bucket_chunk_size: int,
    input_deltas_stats: Dict[int, DeltaStats],
    deltacat_storage=unimplemented_deltacat_storage,
) -> Tuple[List[DeltaAnnotated], int, int]:

    # TODO (pdames): when row counts are available in metadata, use them
    #  instead of bytes - memory consumption depends more on number of
    #  input delta records than bytes.

    # we assume here that we're running on a fixed-size cluster
    # this assumption could be removed, but we'd still need to know the max
    # resources we COULD get for this cluster, and the amount of memory
    # available per CPU should remain fixed across the cluster.
    worker_cpus = int(cluster_resources["CPU"])
    worker_obj_store_mem = float(cluster_resources["object_store_memory"])
    # worker_obj_store_mem = ray_constants.from_memory_units(
    #     cluster_resources["object_store_memory"]
    # )
    if min_pk_index_pa_bytes > 0:
        required_heap_mem_for_dedupe = worker_obj_store_mem - min_pk_index_pa_bytes
        assert required_heap_mem_for_dedupe > 0, (
            f"Not enough required memory available to re-batch input deltas"
            f"and initiate the dedupe step."
        )
        # Size of batched deltas must also be reduced to have enough space for primary
        # key index files (from earlier compaction rounds) in the dedupe step, since
        # they will be loaded into worker heap memory.
        worker_obj_store_mem = required_heap_mem_for_dedupe

    logger.info(f"Total worker object store memory: {worker_obj_store_mem}")
    worker_obj_store_mem_per_task = worker_obj_store_mem / worker_cpus
    logger.info(f"Worker object store memory/task: " f"{worker_obj_store_mem_per_task}")
    worker_task_mem = cluster_resources["memory"]
    logger.info(f"Total worker memory: {worker_task_mem}")
    # TODO (pdames): ensure fixed memory per CPU in heterogenous clusters
    worker_mem_per_task = worker_task_mem / worker_cpus
    logger.info(f"Cluster worker memory/task: {worker_mem_per_task}")

    delta_bytes = 0
    delta_bytes_pyarrow = 0
    delta_manifest_entries = 0
    latest_stream_position = -1
    limited_input_da_list = []

    if input_deltas_stats is None:
        input_deltas_stats = {}

    input_deltas_stats = {
        int(stream_pos): DeltaStats(delta_stats)
        for stream_pos, delta_stats in input_deltas_stats.items()
    }
    for delta in input_deltas:
        manifest = deltacat_storage.get_delta_manifest(delta)
        delta.manifest = manifest
        position = delta.stream_position
        delta_stats = input_deltas_stats.get(delta.stream_position, DeltaStats())
        if delta_stats:
            delta_bytes_pyarrow += delta_stats.stats.pyarrow_table_bytes
        else:
            # TODO (pdames): ensure pyarrow object fits in per-task obj store mem
            logger.warning(
                f"Stats are missing for delta stream position {delta.stream_position}, "
                f"materialized delta may not fit in per-task object store memory."
            )
        manifest_entries = delta.manifest.entries
        delta_manifest_entries += len(manifest_entries)
        for entry in manifest_entries:
            delta_bytes += entry.meta.content_length
            if not delta_stats:
                delta_bytes_pyarrow = delta_bytes * PYARROW_INFLATION_MULTIPLIER
        latest_stream_position = max(position, latest_stream_position)
        if delta_bytes_pyarrow > worker_obj_store_mem:
            logger.info(
                f"Input deltas limited to "
                f"{len(limited_input_da_list)} by object store mem "
                f"({delta_bytes_pyarrow} > {worker_obj_store_mem})"
            )
            break
        delta_annotated = DeltaAnnotated.of(delta)
        limited_input_da_list.append(delta_annotated)

    logger.info(f"Input deltas to compact this round: " f"{len(limited_input_da_list)}")
    logger.info(f"Input delta bytes to compact: {delta_bytes}")
    logger.info(f"Input delta files to compact: {delta_manifest_entries}")
    logger.info(f"Latest input delta stream position: {latest_stream_position}")

    if not limited_input_da_list:
        raise RuntimeError("No input deltas to compact!")

    # TODO (pdames): determine min hash buckets from size of all deltas
    #  (not just deltas for this round)
    min_hash_bucket_count = int(
        max(
            math.ceil(delta_bytes_pyarrow / worker_obj_store_mem_per_task),
            min(worker_cpus, 256),
        )
    )
    logger.info(f"Minimum recommended hash buckets: {min_hash_bucket_count}")

    if hash_bucket_count is None:
        # TODO (pdames): calc default hash buckets from table growth rate... as
        #  this stands, we don't know whether we're provisioning insufficient
        #  hash buckets for the next 5 minutes of deltas or more than enough
        #  for the next 10 years
        hash_bucket_count = min_hash_bucket_count
        logger.info(f"Using default hash bucket count: {hash_bucket_count}")

    if hash_bucket_count < min_hash_bucket_count:
        logger.warning(
            f"Provided hash bucket count ({hash_bucket_count}) "
            f"is less than the min recommended ({min_hash_bucket_count}). "
            f"This compaction job run may run out of memory, or run slowly. To "
            f"resolve this problem either specify a larger number of hash "
            f"buckets when running compaction, omit a custom hash bucket "
            f"count when running compaction, or provision workers with more "
            f"task memory per CPU."
        )

    hash_bucket_chunk_size = user_hash_bucket_chunk_size
    max_hash_bucket_chunk_size = math.ceil(
        worker_obj_store_mem_per_task / PYARROW_INFLATION_MULTIPLIER
    )
    logger.info(f"Max hash bucket chunk size: {max_hash_bucket_chunk_size}")
    if hash_bucket_chunk_size > max_hash_bucket_chunk_size:
        # TODO (pdames): note type of memory to increase (task or object store)
        logger.warning(
            f"Provided hash bucket chunk size "
            f"({user_hash_bucket_chunk_size}) is greater than the max "
            f"recommended ({max_hash_bucket_chunk_size}). This compaction "
            f"job may run out of memory. To resolve this problem either "
            f"specify a smaller hash bucket chunk size when running "
            f"compaction, omit a custom hash bucket chunk size when running "
            f"compaction, or provision workers with more task and object "
            f"store memory per CPU."
        )
    elif not hash_bucket_chunk_size:
        hash_bucket_chunk_size_load_balanced = max(
            math.ceil(max(delta_bytes, delta_bytes_pyarrow) / worker_cpus),
            BYTES_PER_MEBIBYTE,
        )
        hash_bucket_chunk_size = min(
            max_hash_bucket_chunk_size,
            hash_bucket_chunk_size_load_balanced,
        )
        logger.info(f"Default hash bucket chunk size: {hash_bucket_chunk_size}")

    rebatched_da_list = DeltaAnnotated.rebatch(
        limited_input_da_list,
        hash_bucket_chunk_size,
    )

    logger.info(f"Hash bucket chunk size: {hash_bucket_chunk_size}")
    logger.info(f"Hash bucket count: {hash_bucket_count}")
    logger.info(f"Input uniform delta count: {len(rebatched_da_list)}")

    return rebatched_da_list, hash_bucket_count, latest_stream_position



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional

from deltacat.compute.stats.models.manifest_entry_stats import ManifestEntryStats
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.compute.stats.types import StatsType


class DeltaColumnStats(dict):
    """
    Stats container for an individual column of a Delta.
    Provides distinct stats results for each manifest entry of the Delta.

    Example:
        Manifest Entry 1
        =======
        foo bar baz
        A   B   C
        D   E   F

        Manifest Entry 2
        =======
        foo bar baz
        G   H   I
        J   K   L

        DeltaColumnStats("foo",
            ManifestEntryStats([
                StatsResult([A, D]),     #  Manifest Entry 1
                StatsResult([G, J]),     #  Manifest Entry 2
            ]))
        DeltaColumnStats("bar",
            ManifestEntryStats([
                StatsResult([B, E]),     #  Manifest Entry 1
                StatsResult([H, K]),     #  Manifest Entry 2
            ]))
        DeltaColumnStats("baz",
            ManifestEntryStats([
                StatsResult([C, F]),     #  Manifest Entry 1
                StatsResult([I, L]),     #  Manifest Entry 2
            ]))
    """

    @staticmethod
    def of(column: str, manifest_stats: ManifestEntryStats) -> DeltaColumnStats:
        """
        Creates a container of a column name and the column stats for one or more manifest entries.
        """
        dcs = DeltaColumnStats()
        dcs["column"] = column
        dcs["manifestStats"] = manifest_stats

        if manifest_stats:
            # Omit row count for columnar-centric stats
            dcs["stats"] = dcs._merge_manifest_stats()

        return dcs

    @staticmethod
    def build_from_dict(delta_column_stats: List[str, Any]) -> List[DeltaColumnStats]:
        return DeltaColumnStats.of(
            delta_column_stats["column"],
            ManifestEntryStats.build_from_dict(delta_column_stats["manifestStats"]),
        )

    @property
    def column(self) -> str:
        """Returns the column name."""
        return self.get("column")

    @property
    def manifest_stats(self) -> Optional[ManifestEntryStats]:
        """Returns a container that represents stats at the manifest level.

        A container holds a list of computed stats for each manifest entry.
        """
        val: Dict[str, Any] = self.get("manifestStats")
        if val is not None and not isinstance(val, ManifestEntryStats):
            self["manifestStats"] = val = ManifestEntryStats(val)
        return val

    @property
    def stats(self) -> Optional[StatsResult]:
        """Combines the numerical stats for every manifest entry and returns it."""
        val: Dict[str, Any] = self.get("stats")
        if val is not None and not isinstance(val, StatsResult):
            self["stats"] = val = StatsResult(val)
        elif val is None and self.manifest_stats:
            self["stats"] = val = self._merge_manifest_stats()

        return val

    def _merge_manifest_stats(self) -> StatsResult:
        return StatsResult.merge(
            self.manifest_stats.stats, {StatsType.PYARROW_TABLE_BYTES}
        )
